/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2017 All Rights Reserved.
 */
package com.abssqr.plat.biz.shared.dispatcher;

import com.abssqr.plat.common.model.domain.job.JobControl;
import com.abssqr.plat.common.model.domain.task.DispatcherNode;
import com.abssqr.plat.common.model.domain.task.TaskGroup;
import com.abssqr.plat.common.model.repo.job.JobControlRepository;
import com.abssqr.plat.common.util.loggers.CommonLoggers;
import com.general.system.common.model.exception.BaseException;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.ThreadUtil;
import org.apache.ibatis.exceptions.PersistenceException;
import org.assertj.core.util.Lists;
import org.slf4j.Logger;
import org.springframework.dao.CannotAcquireLockException;
import org.springframework.dao.QueryTimeoutException;
import org.springframework.transaction.CannotCreateTransactionException;
import org.springframework.util.CollectionUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import static com.general.system.common.util.MessageUtil.formatMsg;

/**
 * 通用守护线程调度类
 *
 * @author zhenxuan.luo
 * @version $Id: Dispatcher.java, v 0.1 2017-06-30 下午17:45 zhenxuan.luo Exp $
 */
public class Dispatcher<T extends DispatcherNode> implements Runnable {

    private final static Logger LOG = CommonLoggers.COMMON_DISPATCHER;

    private final static int ONE_DAY_SEC = 24 * 60 * 60;

    private String name;

    private int queueSize = 1000;

    private int coreSize = 2;

    private int maxSize = 5;

    private long keepAliveTime = ONE_DAY_SEC;

    private int hungrySize = 10;

    private boolean isRunning = true;

    private long noJobSleepMills = 2000;

    private long queueFullSleepMills = 500;

    private DispatcherLoader<T> dispatcherLoader;

    private DispatcherExecutor<T> dispatcherExecutor;

    private DispatcherSplitor dispatcherSplitor;

    private BlockingQueue<Runnable> queue;

    private ThreadPoolExecutor executor;

    private JobControlRepository jobControlRepository;

    @Override
    public void run() {
        int isEffCnt = 0;
        LOG.info(formatMsg("守护进程{0}启动", name));

        // 1. 初始化队列
        queue = new ArrayBlockingQueue<Runnable>(queueSize);

        // 2. 初始化线程池
        executor = new ThreadPoolExecutor(coreSize, maxSize, keepAliveTime, TimeUnit.SECONDS,
            queue);

        // 3. 注册应用关闭的钩子，尽量保证已调度的任务执行完毕后再关闭应用
        addShutdownHook();

        while (isRunning) {
            try {
                int curQueueSize = queue.size();
                if (curQueueSize < hungrySize) {
                    JobControl jobControl = jobControlRepository.selectByJobName(name);
                    if (jobControl != null && jobControl.getIsEff().isYes()) {
                        isEffCnt = 0;
                        Long start = System.currentTimeMillis();
                        List<T> toDoList = Lists.newArrayList();
                        if (dispatcherSplitor != null) {
                            //一次性获取两个任务组
                            List<TaskGroup> taskGroups = dispatcherSplitor.getTask();
                            if (!CollectionUtils.isEmpty(taskGroups)) {

                                toDoList = dispatcherLoader.loadToDoNodes(new TaskGroupService() {
                                    @Override
                                    public List getTasks() {
                                        return taskGroups.get(0).getTasks();
                                    }

                                    @Override
                                    public void finish() {
                                        dispatcherSplitor.finish(taskGroups.get(0).getId());
                                    }
                                },new TaskGroupService(){

                                    @Override
                                    public List getTasks() {
                                        List spareTasks=new ArrayList();
                                        if(taskGroups.size()>1){
                                            spareTasks=taskGroups.get(1).getTasks();
                                        }
                                        return spareTasks;
                                    }

                                    @Override
                                    public void finish() {
                                        if(taskGroups.size()>1) {
                                            dispatcherSplitor.finish(taskGroups.get(1).getId());
                                        }
                                    }
                                }, queueSize - curQueueSize);
                            }


                        } else {
                            toDoList = dispatcherLoader.loadToDoNodes(queueSize - curQueueSize);
                        }

                        if (CollectionUtils.isEmpty(toDoList)) {
                            ThreadUtil.sleep(noJobSleepMills, LOG);
                            continue;
                        } else {
                            for (T node : toDoList) {
                                final T executeNode = node;
                                executor.execute(() -> dispatcherExecutor.execute(executeNode));
                            }
                        }
                    } else {
                        isEffCnt++;
                        if (isEffCnt > 5) {
                            LOG.warn(formatMsg("定时任务[{0}]执行标记未生效-[{1}]", name, jobControl));
                            isEffCnt = 0;
                        }
                        Thread.sleep(noJobSleepMills);
                        continue;
                    }
                } else {
                    ThreadUtil.sleep(queueFullSleepMills, LOG);
                }
            } catch (BaseException e) {
                LOG.error(formatMsg("守护线程[{0}]执行，遇见异常", name), e);

            } catch (PersistenceException e) {
                LOG.error(formatMsg("守护线程[{0}]执行，数据库链接异常", name), e);
                try {
                    Thread.sleep(3000);
                } catch (InterruptedException e1) {
                    LOG.error(formatMsg("守护线程[{0}]执行，数据库链接异常", name), e);
                }
            } catch (CannotCreateTransactionException e) {
                LOG.error(formatMsg("守护线程[{0}]执行，数据库链接异常", name), e);
                try {
                    Thread.sleep(3000);
                } catch (InterruptedException e1) {
                    LOG.error(formatMsg("守护线程[{0}]执行，数据库链接异常", name), e);
                }
            } catch (QueryTimeoutException e) {
                //mybatis  timeout 抛出
                // LOG.error(formatMsg("守护线程[{0}]执行，数据库链接异常", name), e);
            } catch (CannotAcquireLockException e) {
                //NOWAIT
                LOG.error(formatMsg("守护线程[{0}]执行，CannotAcquireLockException,{1}", name, queue.size()));
                ThreadUtil.sleep(100, LOG);
            } catch (Throwable e) {
                LOG.error(formatMsg("守护线程[{0}]执行，遇见未知异常", name), e);
            }
        }
        LOG.info(formatMsg("守护进程{0}结束", name));

    }

    private void waitJobFinish() {
        isRunning = false;
        boolean isAllFinish = false;
        for (int checkTimes = 0; checkTimes <= 20; checkTimes++) {
            ThreadUtil.sleep(1000, LOG);

            int activeCnt = this.executor.getActiveCount();
            int blockingQueueSize = this.queue.size();

            LogUtil.warn(LOG, "[dispatcherFinish - {0}]活跃线程数[{1}], 缓冲队列任务列表[{2}]", getName(),
                activeCnt, blockingQueueSize);

            if (activeCnt == 0 && blockingQueueSize == 0) {
                isAllFinish = true;
                break;
            }
        }

        if (!isAllFinish) {
            LogUtil.warn(LOG, "[dispatcherFinish - {0}]守护线程队列任务可能未全部结束，需启动恢复方案", getName());
        }
    }

    private void addShutdownHook() {
        Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
            @Override
            public void run() {
                waitJobFinish();
            }
        }));
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public int getQueueSize() {
        return queueSize;
    }

    public void setQueueSize(int queueSize) {
        this.queueSize = queueSize;
    }

    public int getCoreSize() {
        return coreSize;
    }

    public void setCoreSize(int coreSize) {
        this.coreSize = coreSize;
    }

    public int getMaxSize() {
        return maxSize;
    }

    public void setMaxSize(int maxSize) {
        this.maxSize = maxSize;
    }

    public long getKeepAliveTime() {
        return keepAliveTime;
    }

    public void setKeepAliveTime(long keepAliveTime) {
        this.keepAliveTime = keepAliveTime;
    }

    public int getHungrySize() {
        return hungrySize;
    }

    public void setHungrySize(int hungrySize) {
        this.hungrySize = hungrySize;
    }

    public DispatcherLoader<T> getDispatcherLoader() {
        return dispatcherLoader;
    }

    public void setDispatcherLoader(DispatcherLoader<T> dispatcherLoader) {
        this.dispatcherLoader = dispatcherLoader;
    }

    public DispatcherExecutor<T> getDispatcherExecutor() {
        return dispatcherExecutor;
    }

    public void setDispatcherExecutor(DispatcherExecutor<T> dispatcherExecutor) {
        this.dispatcherExecutor = dispatcherExecutor;
    }

    public long getNoJobSleepMills() {
        return noJobSleepMills;
    }

    public void setNoJobSleepMills(long noJobSleepMills) {
        this.noJobSleepMills = noJobSleepMills;
    }

    public long getQueueFullSleepMills() {
        return queueFullSleepMills;
    }

    public void setQueueFullSleepMills(long queueFullSleepMills) {
        this.queueFullSleepMills = queueFullSleepMills;
    }

    public boolean isRunning() {
        return isRunning;
    }

    public void setJobControlRepository(JobControlRepository jobControlRepository) {
        this.jobControlRepository = jobControlRepository;
    }

    public ThreadPoolExecutor getExecutor() {
        return executor;
    }

    public void setExecutor(ThreadPoolExecutor executor) {
        this.executor = executor;
    }

    public DispatcherSplitor getDispatcherSplitor() {
        return dispatcherSplitor;
    }

    public void setDispatcherSplitor(DispatcherSplitor dispatcherSplitor) {
        this.dispatcherSplitor = dispatcherSplitor;
    }
}