/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2020 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.accounting;

import cn.hutool.core.date.DateUtil;
import com.abssqr.plat.common.facade.enums.PlanTypeEnum;
import com.abssqr.plat.common.facade.enums.SyncTaskTypeEnum;
import com.abssqr.plat.common.facade.model.org.OrganizationEntity;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.facade.model.prod.LoanProduct;
import com.abssqr.plat.common.facade.model.rule.AstScreenRule;
import com.abssqr.plat.common.model.domain.accounting.AcctDate;
import com.abssqr.plat.common.model.enums.OrgCodeEnum;
import com.abssqr.plat.common.model.enums.OrgTypeEnum;
import com.abssqr.plat.common.model.exception.AbssqrBizException;
import com.abssqr.plat.common.model.repo.acct.AcctDateRepo;
import com.abssqr.plat.common.model.repo.plan.PlanTransferRuleRepo;
import com.abssqr.plat.common.model.repo.prod.LoanProdRepository;
import com.abssqr.plat.common.util.base.Day;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.VarChecker;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 机构会计日组件
 * @author hanfei
 * @version $Id: OrgAcctDayComponent.java, v 0.1 2020-03-03 4:40 PM hanfei Exp $ 
 */
@Component
public class OrgAcctDayComponent {
    
    @Autowired
    private AcctDateRepo         acctDateRepo;
    @Autowired
    private AstSyncTaskComponent astSyncTaskComponent;
    @Autowired
    private LoanProdRepository   loanProdRepository;
    @Autowired
    private PlanTransferRuleRepo planTransferRuleRepo;
    
    /**
     * 获取机构会计日
     * @param org 机构编码枚举，枚举中含机构编码及类型
     * @return Day 机构会计日
     */
    public Day getAstAcctDay(OrgCodeEnum org){
        VarChecker.checkNotNull(org, "机构编码不能为空");
        AcctDate acctDate = acctDateRepo.getAcctDate(org);
        if (acctDate == null) {
            throw new AbssqrBizException(org.getCode() + "当前机构会计日为空");
        }
        return new Day(acctDate.getAcctDate());
    }
    
    /**
     * 获取转让会计日
     * @return Day 转让会计日
     */
    public Day getTrfAcctDay(){
        AcctDate acctDate = acctDateRepo.getAcctDate(OrgCodeEnum.ABS);
        if (acctDate == null || acctDate.getAcctDate() == null) {
            throw new AbssqrBizException("当前转让系统会计日为空");
        }
        return new Day(acctDate.getAcctDate());
    }
    
    /**
     * 会计日切换
     */
    public void acctDaySwitch(){
        // 取出所有会计日
        List<AcctDate> acctDates = acctDateRepo.queryAcctDate();
        VarChecker.checkNotEmpty(acctDates, "获取所有会计日为空");
        // 按照机构类型分组
        Map<OrgTypeEnum, List<AcctDate>> acctDateMap = acctDates.stream().collect(
                Collectors.groupingBy(one -> one.getOrgCode().getType()));
    
        // 取AST类型会计日循环
        List<AcctDate> astDays = acctDateMap.get(OrgTypeEnum.AST);
        VarChecker.checkNotEmpty(astDays, "获取资产机构会计日为空");
        List<AcctDate> trfDays = acctDateMap.get(OrgTypeEnum.ABS);
        VarChecker.checkNotEmpty(trfDays, "获取转让会计日为空");
        VarChecker.checkArgument(trfDays.size() == 1, "转让会计日为多条逻辑错误");
        AcctDate trfDay = trfDays.get(0);
        
        for (AcctDate astDay : astDays){
            /**
             * 跳出不执行场景：
             * 1、eff 为0=false
             * 2、机构会计日大于等于转让会计日
             * 3、当前机构会计日模型转换未完成
             */
            if (!astDay.getEff()
                || !DateTimeUtil.isBeforeOnDate(astDay.getAcctDate(), trfDay.getAcctDate())
                || !astSyncTaskComponent.isTaskFinishByOrgCode(astDay.getAcctDate(), astDay.getOrgCode(), SyncTaskTypeEnum.AST_MOD)){
                // 跳出等待下一次调度
                continue;
            }
            
            // 当前机构会计日+1
            Date nextDate = DateTimeUtil.addDays(astDay.getAcctDate(), 1);
            // 判断+1后日期对应机构编码的数据同步是否完成
            if (astSyncTaskComponent.isTaskFinishByOrgCode(nextDate, astDay.getOrgCode(), SyncTaskTypeEnum.AST_SYNC)) {
                // 如果已完成抛异常
                throw new AbssqrBizException(MessageUtil.formatMsg(
                        "error-[{0}]机构会计日{1}的模型转换已完成不能进行切换操作，请人工确认",
                        astDay.getOrgCode().getCode(), DateUtil.formatDate(nextDate)));
            }
    
            // 判断机构数据五矿准备完成标记
            if (!astSyncTaskComponent.isWKDataReady(nextDate, astDay.getOrgCode())) {
                // 未完成直接跳出等待下一次调度 判断知道五矿数据准备完毕
                continue;
            }
    
            // 更新机构会计日+1
            astDay.setAcctDate(nextDate);
            acctDateRepo.modifyAcctDate(astDay);
        }
        
        // 判断转让会计日是否可执行
        if (!trfDay.getEff()){
            return;
        }
        // 循环更新后的机构会计日
        for (AcctDate astDay : astDays){
            // 如果机构会计日与当前转让会计日为同一天设置标记可以更新
            if (DateTimeUtil.isSameDay(astDay.getAcctDate(), trfDay.getAcctDate())){
                // 更新转让会计日+1
                trfDay.setAcctDate(DateTimeUtil.addDays(trfDay.getAcctDate(), 1));
                acctDateRepo.modifyAcctDate(trfDay);
                break;
            }
        }
    }
    
    
    /**
     * 获取项目计划对应的机构编码
     * @param planBase
     * @return
     */
    public OrgCodeEnum getOrgCode(PlanBase planBase){
        String orgCode = "";
        if (planBase.getPlanType() == PlanTypeEnum.SPV && planBase instanceof OrganizationEntity) {
            orgCode = ((OrganizationEntity) planBase).getManagerOrgCode();
        } else {
            // 暂时获取 计划的发起机构
            PlanTransferRule transferRule = planTransferRuleRepo.getByPlanNo(planBase.getPlanNo());
            // 筛选规则
            List<AstScreenRule> astScreenRuleList = transferRule.getAstScreenRuleList();
            VarChecker.checkNotEmpty(astScreenRuleList, "未找到对应计划的筛选规则={0}", planBase.getPlanNo());
            List<LoanProduct> products = loanProdRepository.getBybssqrProdCodes(
                    astScreenRuleList.stream().map(AstScreenRule::getAbssqrProdCode).collect(Collectors.toList()));
            VarChecker.checkNotEmpty(products, "未找到对应计划的产品信息={0}", planBase.getPlanNo());
            // 获取管理机构编码
            List<String> managerCodes = products.stream().map(LoanProduct::getManagerOrgCode).collect(Collectors.toList());
            VarChecker.checkNotEmpty(managerCodes, "未找到对应计划的管理机构编码={0}", planBase.getPlanNo());
            orgCode = managerCodes.get(0);
        }
        OrgCodeEnum orgCodeEnum = OrgCodeEnum.getByCode(orgCode);
        VarChecker.checkNotNull(orgCodeEnum, "{0}计划对应的机构编码不匹配{1}", planBase.getPlanNo(), orgCode);
        return orgCodeEnum;
    }
    
}
