/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2020 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.ast;

import com.abssqr.plat.common.facade.enums.LoanStatusEnum;
import com.abssqr.plat.common.facade.enums.OwnStatusEnum;
import com.abssqr.plat.common.facade.enums.TaskStatusEnum;
import com.abssqr.plat.common.model.domain.ast.AstConditions;
import com.abssqr.plat.common.model.domain.ast.AstPoolEntity;
import com.abssqr.plat.common.model.domain.ast.Loan;
import com.abssqr.plat.common.model.domain.ast.LoanAstGroup;
import com.abssqr.plat.common.model.domain.contract.LoanContract;
import com.abssqr.plat.common.model.repo.pool.AstPoolRepository;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Maps;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.Map;

/**
 * 资产池资产处理
 * @author hanfei
 * @version $Id: AstPoolManager.java, v 0.1 2020-02-13 11:23 PM hanfei Exp $ 
 */
@Component
public class AstPoolManager {
    
    @Autowired
    private AstPoolRepository astPoolRepository;
    
    /**
     * 交易后资产池资产处理
     * @param loan
     */
    public void astPoolItemTrans(Loan loan){
        OwnStatusEnum ownStatus = loan.getAstOrg().getAstOwnership().getOwnStatus();
        if(ownStatus == OwnStatusEnum.PARTIAL_TRANS_OUT){
            astPoolRepository.updateSlice(loan.getAstNo(), loan.getTermCnt());
        }else if(ownStatus == OwnStatusEnum.TRANS_OUT){
            astPoolRepository.delete(loan.getAstNo(), loan.getTermCnt());
        }
    }
    
    /**
     * 删除资产池资产
     * @param astNo
     * @param termCnt
     */
    public void removeAstPoolItem(String astNo, Integer termCnt){
        astPoolRepository.delete(astNo, termCnt);
    }
    
    /**
     * 构建资产池资产
     * @param loan
     */
    public void buildAstPoolEntity(Loan loan, boolean isSlice, Date bizDate) {
        
        if (loan.getStatus() == LoanStatusEnum.CLEAR || loan.getStatus() == LoanStatusEnum.REVERSE) {
            return;
        }
        LoanContract loanContract = loan.getContract();
        VarChecker.checkNotNull(loanContract, "未组装贷款合同[{0}]", loan.getAstNo());
        
        AstPoolEntity astPoolEntity = new AstPoolEntity();
        astPoolEntity.setAstNo(loan.getAstNo());
        astPoolEntity.setManagerOrgCode(loan.getAstOrg().getManagerOrgCode());
        astPoolEntity.setTermNo(loan.getTermCnt());
        astPoolEntity.setStartDate(loan.getAstPeriod().getStartDate());
        astPoolEntity.setEndDate(loan.getAstPeriod().getEndDate());
        astPoolEntity.setBizDate(bizDate);
        astPoolEntity.setTaskStatus(TaskStatusEnum.INIT);
        astPoolEntity.setPrinBal(loan.getBal().getAllPrin());
        astPoolEntity.setOwnOrg(loan.getAstOrg().getAstOwnership().getOwnOrg());
        // 切片标记
        astPoolEntity.setSlice(isSlice);
        Map<String, String> context = Maps.newHashMap();
        context.put(AstConditions.CERT_NO, loanContract.getLoanCustInfo().getCertNo());
        context.put(AstConditions.CERT_TYPE, loanContract.getLoanCustInfo().getCertType().getCode());
        context.put(AstConditions.INTER_RATE, loanContract.getLoanProperty().getRate().toString());
        context.put(AstConditions.RATE_TYPE, loanContract.getLoanProperty().getRateType().getCode());
        astPoolEntity.setContext(context);
        
        astPoolRepository.create(astPoolEntity);
    }
    
    /**
     * 回购场景，判断机构资产是否都为放款池机构，如果是部分回购则创建切片资产
     * @param loan
     */
    public void buildAstPoolByBuyBack(Loan loan, Date bizDate) {
        
        if (loan.getStatus() == LoanStatusEnum.CLEAR || loan.getStatus() == LoanStatusEnum.REVERSE) {
            return;
        }
        boolean exist = false;
        // 判断资产池中是否已有资产
        AstPoolEntity entity = astPoolRepository.getAstPoolItem(
                loan.getAstNo(), loan.getAstOrg().getAstOwnership().getLenderOrgCode());
        if (entity != null){
            exist = true;
        }
        
        LoanContract loanContract = loan.getContract();
        VarChecker.checkNotNull(loanContract, "未组装贷款合同[{0}]", loan.getAstNo());
        // 获取可作为资产池资产的 放款池机构资产
        LoanAstGroup loanAstGroup = loan.filterLoanAstByOrg(
                loan.getAstOrg().getAstOwnership().getLenderOrgCode(), loan.getTermCnt());
        
        if (exist){
            // isAllAst 为false 更新资产标记切片。
            if (! loanAstGroup.isAllAst()){
                astPoolRepository.updateSlice(loan.getAstNo(), loan.getTermCnt());
            }
        }else{
            // isAllAst 为 true 创建完整资产。 为false 创建资产标记切片。
            this.buildAstPoolEntity(loan, ! loanAstGroup.isAllAst(), bizDate);
        }

//            // 根据机构资产创建待选切片资产
//            for (LoanAst loanAst : loanAstGroup.getLoanAsts()){
//                if (loanAst.getPrinBal().isZero()){
//                    continue;
//                }
//                // 分期
//                LoanInstalDomain instalDomain = loanAst.getRefInstal();
//                AstPoolEntity astPoolEnity = new AstPoolEntity();
//                astPoolEnity.setAstNo(instalDomain.getAstNo());
//                astPoolEnity.setManagerOrgCode(instalDomain.getAstOrg().getManagerOrgCode());
//                astPoolEnity.setTermNo(instalDomain.getTermNo());
//                astPoolEnity.setStartDate(instalDomain.getAstPeriod().getStartDate());
//                astPoolEnity.setEndDate(instalDomain.getAstPeriod().getEndDate());
//                astPoolEnity.setBizDate(SystemDayUtil.getDay().getStandardDate());
//                astPoolEnity.setTaskStatus(TaskStatusEnum.INIT);
//                astPoolEnity.setPrinBal(loanAst.getPrinBal());
//                astPoolEnity.setOwnOrg(loanAst.getOwnOrg());
//                Map<String, String> context = Maps.newHashMap();
//                context.put(AstConditions.CERT_NO, loanContract.getLoanCustInfo().getCertNo());
//                context.put(AstConditions.CERT_TYPE, loanContract.getLoanCustInfo().getCertType().getCode());
//                context.put(AstConditions.INTER_RATE, loanContract.getLoanProperty().getRate().toString());
//                context.put(AstConditions.RATE_TYPE, loanContract.getLoanProperty().getRateType().getCode());
//                astPoolEnity.setContext(context);
//                astPoolRepository.create(astPoolEnity);
//            }
    
    }
}
