/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.core.service.plan;

import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.PlanAllocStatusEnum;
import com.abssqr.plat.common.facade.model.plan.PlanAcctTitle;
import com.abssqr.plat.common.facade.model.plan.PlanAlloc;
import com.abssqr.plat.common.facade.model.plan.PlanAllocItem;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.utils.CalcUtil;
import com.abssqr.plat.common.model.domain.plan.PlanAcctTitleLog;
import com.abssqr.plat.common.model.domain.plan.PlanAcctTitleTransDetail;
import com.abssqr.plat.common.model.repo.plan.PlanAcctTitleLogRepo;
import com.general.system.bpm.enums.ApproveStatusEnum;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.MoneyUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 分配试算
 *
 * @author zhenxuan.luo
 * @version com.abssqr.plat.core.service.plan: PlanAllocAmtCalcManager.java, v 0.1 2019-06-04 11:25 zhenxuan.luo Exp $
 */
@Component
public class PlanAllocAmtCalcManager {

    private final static Logger LOGGER = LoggerFactory.getLogger(PlanAllocAmtCalcManager.class);

    @Autowired
    private PlanSettleManager planSettleManager;

    @Autowired
    private PlanAcctTitleLogRepo planAcctTitleLogRepo;

    /**
     * 一个分配计划 根据金额计算分配计划分摊
     *
     * @param planEntity
     * @param allocAmt
     * @param planAlloc
     * @param planAcctTitles
     */
    public void calcCfrmAmtByAmt(PlanBase planEntity, Money allocAmt, PlanAlloc planAlloc,
                                 List<PlanAcctTitle> planAcctTitles) {

        // 填入息费试算金额
        List<PlanAcctTitle> clonedAcctTitles = Lists.newArrayList();

        // 备份：计算过程中会对金额发生改变，防止误更新
        for (PlanAcctTitle planAcctTitle : planAcctTitles) {
            clonedAcctTitles.add(planAcctTitle.clone());
        }

        // 试算费用
        this.calcFeeAllocAmtForItems(planAlloc.getClearDate(), planEntity, planAlloc, planAcctTitles, clonedAcctTitles,
            true);

        // 填入本金试算金额
        this.calcAllocInvestAmtForItems(planAlloc, planAcctTitles, clonedAcctTitles, true);

        // 待分配金额 不大于零 直接跳过分摊
        if (!allocAmt.isGreaterThanZero()){
            return;
        }
        Money remainAmt = allocAmt.clone();
        // 进行分摊
        for (PlanAllocItem planAllocItem : planAlloc.getPlanAllocItemList()) {
            if (remainAmt.isGreaterThanZero()) {
                // 重置确认金额
                planAllocItem.setConfirmTaxAmt(new Money());
                planAllocItem.setConfirmAmt(new Money());
                
                // 试算金额及试算增值税 合计
                Money sumAmt = planAllocItem.getPayWithOutTaxAmt().clone();
                if (planAllocItem.getTax()) {
                    sumAmt.addTo(planAllocItem.getTrialTaxAmt());
                }
                if (MoneyUtil.compare(remainAmt, sumAmt) >= 0) {
                    planAllocItem.setConfirmAmt(planAllocItem.getTrialAmt());
                    if (planAllocItem.getTax()) {
                        // 如果分摊金额足够，则直接分摊
                        planAllocItem.setConfirmTaxAmt(planAllocItem.getTrialTaxAmt());
                    }
                } else {
                    // 不含税的分摊，有多少摊多少
                    if (!planAllocItem.getTax()) {
                        planAllocItem.setConfirmAmt(remainAmt);
                    } else {
                        // 根据是否支付判断 ， 目前金融服务费会设置不支付标记
                        if(planAllocItem.getPayTrial()) {
                            Money disbCfrmAmt = CalcUtil.disbAmtWithTax(remainAmt, planEntity.getVatRate());
                            Money disbCfrmTaxAmt = remainAmt.subtract(disbCfrmAmt);
    
                            planAllocItem.setConfirmAmt(disbCfrmAmt);
                            planAllocItem.setConfirmTaxAmt(disbCfrmTaxAmt);
                        }else{
                            planAllocItem.setConfirmTaxAmt(remainAmt);
                            Money confirmAmt = CalcUtil.calcVatConfirmForNet(remainAmt, planEntity.getVatRate());
                            planAllocItem.setConfirmAmt(confirmAmt);
                        }
                    }
                }
                //分配金额
                remainAmt = remainAmt.subtract(planAllocItem.getPayWithOutTaxAmt().add(planAllocItem.getConfirmTaxAmt()));
            }
        }
    }

    /**
     * 按信托计划上配置的分配时间和顺序，试算分配计划
     *
     * @param planEntity
     * @param planAllocs
     * @param planAcctTitles
     */
    public void calcTriAmtByScheduler(PlanBase planEntity, List<PlanAlloc> planAllocs,
                                      List<PlanAcctTitle> planAcctTitles, boolean isSetTrial) {

        List<PlanAcctTitle> clonedAcctTitles = Lists.newArrayList();

        // 备份：计算过程中会对金额发生改变，防止误更新
        for (PlanAcctTitle planAcctTitle : planAcctTitles) {
            clonedAcctTitles.add(planAcctTitle.clone());
        }

        // 顺序遍历：模拟结息费，填充金额
        for (PlanAlloc planAlloc : planAllocs) {
            Date simuAcctDate = planAlloc.getClearDate();
            // 费用试算
            calcFeeAllocAmtForItems(simuAcctDate, planEntity, planAlloc, clonedAcctTitles, planAcctTitles, isSetTrial);
        }

        // 逆序遍历：填本金
        for (int i = planAllocs.size() - 1; i >= 0; i--) {
            PlanAlloc planAlloc = planAllocs.get(i);
            // 本金试算
            calcAllocInvestAmtForItems(planAlloc, clonedAcctTitles, planAcctTitles, isSetTrial);
        }
    }

    /**
     * 试算费用
     *
     * @param accountDate
     * @param planEntity
     * @param planAlloc
     * @param clonedAcctTitleList
     * @param originalAcctTitleList
     * @param isSetTrial
     */
    private void calcFeeAllocAmtForItems(Date accountDate, PlanBase planEntity, PlanAlloc planAlloc,
                                         List<PlanAcctTitle> clonedAcctTitleList,
                                         List<PlanAcctTitle> originalAcctTitleList, boolean isSetTrial) {

        accountDate = DateTimeUtil.getStandardDate(accountDate);

        // 当分配计划状态不为 未分配 或审批状态 已经通过 的不需要进行试算
        if (planAlloc.getStatus() != PlanAllocStatusEnum.UNASSIGNE
            || planAlloc.getApproveStatus() == ApproveStatusEnum.PASSED) {
            return;
        }

        Map<String, PlanAcctTitle> planAcctTitleMap = clonedAcctTitleList.stream().collect(
            Collectors.toMap(PlanAcctTitle::getTitleNo, item -> item));

        Map<String, PlanAcctTitle> orignalPlanAcctTitleMap = originalAcctTitleList.stream().collect(
            Collectors.toMap(PlanAcctTitle::getTitleNo, item -> item));

        List<PlanAcctTitleTransDetail> titleTransDtls = planSettleManager
            .settle(planEntity, accountDate, clonedAcctTitleList, false);

        Map<String, PlanAcctTitleTransDetail> transDetailMap = Maps.newHashMap();
        titleTransDtls.forEach(item -> transDetailMap.put(item.getPlanAcctTitle().getTitleNo(), item));

        for (PlanAllocItem planAllocItem : planAlloc.getPlanAllocItemList()) {

            PlanAcctTitle planAcctTitle = planAcctTitleMap.get(planAllocItem.getTitleNo());
            VarChecker.checkNotNull(planAcctTitle, "分配方案对应的分配事项不存在: [{0}]", planAllocItem);

            if (planAcctTitle.getTitleType().isInvest()) {
                continue;
            }

            Money settleAmt;
            PlanAcctTitleTransDetail transDetail = transDetailMap.get(planAcctTitle.getTitleNo());

            if (null == transDetail) {
                Date settleDay = planSettleManager.getSettleDay(planEntity.getEffectDate(), accountDate);
                if (settleDay == null) {
                    settleAmt = new Money();
                } else {
                    // 未找到结息记录，只会因为当前会计日期小于科目的最近结息日期，说明在当前会计日期已经结过息，取当时的结息记录。
                    // 可能会存在，结过息之后又发起兑付的情况，属于极端场景，理论上上这种情况不应该再走到这里
                    List<PlanAcctTitleLog> planAcctTitleLogs = planAcctTitleLogRepo.getByEventAndTitleNo(planEntity.getPlanNo(),
                        planAcctTitle.getTitleNo(), settleDay, EventCodeEnum.TITLE_SETTLE);
                    VarChecker.checkArgument(CollectionUtils.isNotEmpty(planAcctTitleLogs), "科目[{0}]在[{1}-{2}]找不到结息记录",
                        planAcctTitle.getTitleName(), DateTimeUtil.getStandardDate(settleDay),planAcctTitle.getRefNo());
                    VarChecker.checkArgument(planAcctTitleLogs.size() == 1, "科目[{0}]在[{1}-{2}]找到多条结息记录",
                        planAcctTitle.getTitleName(), DateTimeUtil.getStandardDate(settleDay),planAcctTitle.getRefNo());

                    PlanAcctTitleLog planAcctTitleLog = planAcctTitleLogs.get(0);
                    settleAmt = planAcctTitleLog.getAfBal();
                }
            } else {
                settleAmt = planAcctTitle.getBal();
            }

            Money trialAmt = planAcctTitle.pay(settleAmt);
            Money curPayAmt;

            if (transDetail == null || transDetail.isUpdate()) {
                // 实际结息
                curPayAmt = trialAmt;
            } else {
                // 不用结息的科目
                PlanAcctTitle originalPlanAcctTitle = orignalPlanAcctTitleMap.get(planAllocItem.getTitleNo());
                curPayAmt = originalPlanAcctTitle.getBal();
            }

            planAllocItem.setCurrPay(curPayAmt);
            if (isSetTrial) {
                planAllocItem.setTrialAmt(settleAmt);
                if (planAllocItem.getTax()) {
                    planAllocItem
                        .setTrialTaxAmt(
                            CalcUtil.calcVatTaxForNet(planAllocItem.getTrialAmt(), planEntity.getVatRate()));
                }
            }
        }
    }

    /**
     * 填入本金试算金额
     *
     * @param planAlloc
     * @param clonedAcctTitleList
     * @param originalAcctTitleList
     * @param isSetTrial
     */
    private void calcAllocInvestAmtForItems(PlanAlloc planAlloc,
                                            List<PlanAcctTitle> clonedAcctTitleList,
                                            List<PlanAcctTitle> originalAcctTitleList, boolean isSetTrial) {

        Map<String, PlanAcctTitle> clonedPlanAcctTitleMap = clonedAcctTitleList.stream().collect(
            Collectors.toMap(PlanAcctTitle::getTitleNo, item -> item));

        Map<String, PlanAcctTitle> originaldPlanAcctTitleMap = originalAcctTitleList.stream().collect(
            Collectors.toMap(PlanAcctTitle::getTitleNo, item -> item));

        for (PlanAllocItem planAllocItem : planAlloc.getPlanAllocItemList()) {

            PlanAcctTitle planAcctTitle = clonedPlanAcctTitleMap.get(planAllocItem.getTitleNo());
            VarChecker.checkNotNull(planAcctTitle, "分配方案对应的分配事项不存在: [{0}]", planAllocItem);

            if (!planAcctTitle.getTitleType().isInvest()) {
                continue;
            }
            PlanAcctTitle originalTitle = originaldPlanAcctTitleMap.get(planAllocItem.getTitleNo());

            planAllocItem.setCurrPay(originalTitle.getBal());

            if (isSetTrial) {
                Money amt = planAcctTitle.pay(planAcctTitle.getBal());
                planAllocItem.setTrialAmt(amt);
            }
        }
    }
}