/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2019 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.plan;

import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.model.domain.clearing.PlanClearingLog;
import com.abssqr.plat.common.model.domain.clearing.PlanClearingLogKey;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalTransDetail;
import com.abssqr.plat.common.model.repo.clearing.PlanClearingLogRepo;
import com.abssqr.plat.common.util.profiler.PerformTrace;
import com.general.system.common.model.Money;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * 清算流水
 * @author zhenxuan.luo
 * @version com.abssqr.plat.core.service.plan: PlanClearingManager.java, v 0.1 2019-07-12 15:11 zhenxuan.luo Exp $ 
 */
@Component
public class PlanClearingManager {
    
    @Autowired
    private PlanClearingLogRepo planClearingLogRepo;
    
    /**
     * 构造清算
     * 包含：
     *      转让清算
     *      还款补偿清算
     *      还款清算 及 应付利息的清算
     *      资产转让清算
     *      放款冲正清算
     * @param instalTransDetails
     * @param transCode
     * @param clearAccountDate
     * @param filter
     */
    @PerformTrace
    public void processClearing(List<LoanInstalTransDetail> instalTransDetails, TransCodeEnum transCode,
                                Date clearAccountDate, ClearingAmtFilter filter) {
        
        Map<PlanClearingLogKey, PlanClearingLog> clearingLogMap = Maps.newHashMap();
        
        for (LoanInstalTransDetail instalTransDetail : instalTransDetails) {
            LoanInstalLog log = instalTransDetail.getActiveLog();
            
            PlanClearingLogKey clearingLogKey = new PlanClearingLogKey(log.getOwnOrg(), log.getLenderOrgCode(),
                    transCode);
            PlanClearingLog clearingLog = clearingLogMap.get(clearingLogKey);
            if (null == clearingLog) {
                clearingLog = buildClearingLog(log.getOwnOrg(), log.getLenderOrgCode(),
                        transCode, clearAccountDate,
                        log.getSeqNo(), log.getBizNo(), log.getExecTms());
                        
                clearingLogMap.put(clearingLogKey, clearingLog);
            }
            clearingLog.getTransAmt().addTo(filter.getAmt(instalTransDetail));
        }
        
        List<PlanClearingLog> planClearingLogs = Lists.newArrayList();
        for (PlanClearingLog clearingLog : clearingLogMap.values()) {
            if (clearingLog.getTransAmt().isGreaterThanZero()) {
                planClearingLogs.add(clearingLog);
            }
        }
        planClearingLogRepo.createBatch(planClearingLogs);
    }
    
    @PerformTrace
    public void processClearingForLoanAst(List<LoanInstalTransDetail> instalTransDetails, TransCodeEnum transCode,
                                Date clearAccountDate, ClearingFilter filter) {
        
        Map<PlanClearingLogKey, PlanClearingLog> clearingLogMap = Maps.newHashMap();
        
        for (LoanInstalTransDetail instalTransDetail : instalTransDetails) {
            LoanInstalLog log = instalTransDetail.getActiveLog();
            Map<PlanClearingLogKey, Money> clearingAmt = filter.getAmt(instalTransDetail);
            
            for (PlanClearingLogKey key : clearingAmt.keySet()) {
                Money amt = clearingAmt.get(key);
    
                PlanClearingLog clearingLog = clearingLogMap.get(key);
                if (null == clearingLog) {
                    clearingLog = buildClearingLog(key.getPlanNo(), key.getCntPtyCode(),
                            transCode, clearAccountDate,
                            log.getSeqNo(), log.getBizNo(), log.getExecTms());
        
                    clearingLogMap.put(key, clearingLog);
                }
                clearingLog.getTransAmt().addTo(amt);
            }
        }
        
        List<PlanClearingLog> planClearingLogs = Lists.newArrayList();
        for (PlanClearingLog clearingLog : clearingLogMap.values()) {
            if (clearingLog.getTransAmt().isGreaterThanZero()) {
                planClearingLogs.add(clearingLog);
            }
        }
        planClearingLogRepo.createBatch(planClearingLogs);
    }
    
    private PlanClearingLog buildClearingLog(String planNo, String cntPtyCode,
                                            TransCodeEnum transCode, Date AccountDate,
                                            String seqNo, String bizNo, String execTms) {
        PlanClearingLog clearingLog = new PlanClearingLog();
        clearingLog.setPlanNo(planNo);
        clearingLog.setCntPtyCode(cntPtyCode);
        clearingLog.setTransCode(transCode);
        clearingLog.setAccountDate(AccountDate);
        clearingLog.setSeqNo(seqNo);
        clearingLog.setBizNo(bizNo);
        clearingLog.setExecTms(execTms);
        return clearingLog;
    }
    
    /**
     * 清算金额扩展接口
     */
    public interface ClearingAmtFilter {
        Money getAmt(LoanInstalTransDetail instalTransDetail);
    }
    
    /**
     * 清算金额扩展接口
     */
    public interface ClearingFilter {
        Map<PlanClearingLogKey, Money> getAmt(LoanInstalTransDetail instalTransDetail);
    }
}