/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.core.service.repo;

import com.abssqr.plat.common.facade.base.DebtMoney;
import com.abssqr.plat.common.facade.enums.DebtMoneyFieldEnum;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.LoanAstStatusEnum;
import com.abssqr.plat.common.facade.enums.PlanTypeEnum;
import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.facade.model.org.OrganizationEntity;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.utils.CalcUtil;
import com.abssqr.plat.common.model.domain.ast.*;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstProfitLog;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstProfitTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanAstTransDetail;
import com.abssqr.plat.common.model.domain.loanTrans.LoanInstalTransDetail;
import com.abssqr.plat.common.model.utils.DecimalCalcUtil;
import com.abssqr.plat.common.util.profiler.PerformTrace;
import com.abssqr.plat.common.util.utils.BigDecimalUtil;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.service.plan.calc.PlanHandler;
import com.abssqr.plat.core.service.profit.ProfitSplitCalculator;
import com.abssqr.plat.core.service.profit.ProfitSplitComponent;
import com.abssqr.plat.core.share.context.PlanContextHolder;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.MoneyUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 资产流出：还款，减免等操作，还款金额分摊
 * @author zhenxuan.luo
 * @version com.abssqr.plat.core.service.repo: AstRepoComponent.java, v 0.1 2019-11-20 上午24:51 zhenxuan.luo Exp $
 */
@Component
public class AstFlowOutComponent {

    private final static Logger LOGGER = LoggerFactory.getLogger(AstFlowOutComponent.class);

    private static List<TransCodeEnum> allowedTransCodes = Lists.newArrayList(TransCodeEnum.REPO, TransCodeEnum.EXMPT,
        TransCodeEnum.AST_SIMU, TransCodeEnum.GRANT_REVERSE);

    @Autowired
    private PlanManager planManager;

    @Autowired
    private AstFlowOutComponent astFlowOutComponent;

    @Autowired
    private ProfitSplitComponent profitSplitComponent;

    @Autowired
    ProfitSplitCalculator profitSplitCalculator;

    /**
     * 还款操作资产流出
     * 摊本金、摊利息
     * @param instalTrans
     * @param transAmt
     * @param fundAmt
     */
    @PerformTrace
    public void flowOutInstal(LoanInstalTransDetail instalTrans, DebtMoney transAmt, DebtMoney fundAmt,
                              boolean forceSplit) {

        EventCodeEnum eventCode = instalTrans.getActiveLog().getEventCode();
        VarChecker.checkArgument(allowedTransCodes.contains(eventCode.getSceneCode()),
            "资产[{0}-{1}]流出处理：只允许资产流出交易（还款，减免）进行流出处理。传入事件为[{2}]",
            instalTrans.getLoanInstal().getAstNo(), instalTrans.getLoanInstal().getAstNo(),
            instalTrans.getActiveLog().getEventCode());

        // 1. 摊本金
        if (transAmt.getByFieldType(DebtMoneyFieldEnum.PRIN).isGreaterThanZero()) {
            repoPrin(instalTrans, transAmt.getByFieldType(DebtMoneyFieldEnum.PRIN),
                    instalTrans.getActiveLog().getEventCode());
        }

        // 2. 摊利息 and 费用
        repoIntr(instalTrans, transAmt, fundAmt, instalTrans.getActiveLog().getEventCode(), forceSplit);
    }

    /**
     * 本金分摊
     *
     * @param instalTrans
     * @param prinAmt
     * @param eventCode
     */
    private  Map<String, Money> repoPrin(LoanInstalTransDetail instalTrans, Money prinAmt, EventCodeEnum eventCode) {

        if (prinAmt.isZero()) {
            return null;
        }
        LoanInstalDomain instalDomain = instalTrans.getLoanInstal();
        List<LoanAst> loanAsts = instalDomain.getLoanAsts();
        Map<String, LoanAstTransDetail> astTransDetailMap = Maps.newHashMap();

        //计算本金摊销
        Map<String, Money> transAmtMap = calcRepoPrin(prinAmt, instalDomain);

        //4 摊销入账
        for (LoanAst loanAst : loanAsts) {
            Money transAmt = transAmtMap.get(loanAst.getLoanAstNo());
            if (transAmt == null || transAmt.isZero()) {
                continue;
            }
            // 4.1 构造交易明细
            LoanAstTransDetail astTransDetail = LoanAstTransDetail.createTransDtl(loanAst, instalTrans, eventCode);
            astTransDetailMap.put(loanAst.getLoanAstNo(), astTransDetail);
            // 4.2 入账
            astTransDetail.processTransAmt(transAmt);
        }

        // 5 校验
        Money transDetailAmtSum = new Money();
        for (LoanAstTransDetail astTransDetail : astTransDetailMap.values()) {
            LoanAst loanAst = astTransDetail.getLoanAst();
            LoanAstLog loanAstLog = astTransDetail.getActiveLog();

            Money amt = loanAstLog.getPrinAmt();
            transDetailAmtSum.addTo(amt);

            if (eventCode.getSceneCode() == TransCodeEnum.REPO) {
                if (checkIsClear(loanAst.getOwnOrg(), loanAstLog.getAccountDate(), loanAst.getProdNo())) {
                    loanAstLog.getClrgAmt().addTo(amt);
                }
            }
        }

        VarChecker.checkEquals(transDetailAmtSum, prinAmt,
            "资产[{0}-{1}]流出处理：本金分配汇总[{2}]和待分配金额[{3}]不一致",
            instalTrans.getLoanInstal().getAstNo(), instalTrans.getLoanInstal().getAstNo(),
            transDetailAmtSum, prinAmt);
        return transAmtMap;
    }

    /**
     * 计算本金摊销
     * @param prinAmt
     * @param instalDomain
     * @return
     */
    public Map<String, Money> calcRepoPrin(Money prinAmt, LoanInstalDomain instalDomain) {
        Money remain = prinAmt.clone();
        List<LoanAst> loanAsts = instalDomain.getLoanAsts();
        //重新排个序，防止以后尾数处理逻辑优化，依赖loanAst顺序，导致前后分摊结果不一致
        Collections.sort(loanAsts,Comparator.comparing(LoanAst::getLoanAstNo));
        Map<String, BigDecimal> ratios = Maps.newHashMap();
        Map<String, BigDecimal> dist = Maps.newHashMap();

        Map<String, LoanAst> loanAstMap = Maps.newHashMap();
        // 1. 构造本金比例
        for (LoanAst loanAst : loanAsts) {
            if (loanAst.getAstStatus() != LoanAstStatusEnum.NORMAL) {
                continue;
            }
            dist.put(loanAst.getLoanAstNo(), loanAst.getPrinBal().getAmount());
            loanAstMap.put(loanAst.getLoanAstNo(), loanAst);
        }

        DecimalCalcUtil.calcSplitFactor(ratios, dist, false);

        // 2. 计算各机构资产本金摊销分配
        //记录每个机构资产本金摊销值
        Map<String, Money> transAmtMap = Maps.newHashMap();

        for (LoanAst loanAst : loanAsts) {
            if (loanAst.getAstStatus() != LoanAstStatusEnum.NORMAL) {
                continue;
            }
            //计算摊销
            BigDecimal ratio = ratios.get(loanAst.getLoanAstNo());
            Money entryAmt = CalcUtil.calcMoneyRatio(prinAmt, ratio);
            Money transAmt = entryAmt(remain, entryAmt, loanAst.getPrinBal());
            transAmtMap.put(loanAst.getLoanAstNo(), transAmt);
            if (remain.isZero()) {
                break;
            }
        }
        // 3. 尾数处理
        if (remain.isGreaterThanZero()) {
            // 计算 已还/总本金
            // 根据已还本金/总本金的比例排序，越小说明本金还的越少，优先摊销
            List<Pair<String, BigDecimal>> outRatioPairs = Lists.newArrayList();
            transAmtMap.forEach((loanAstNo, transAmt) -> {
                BigDecimal totalBal = dist.get(loanAstNo);
                BigDecimal outAmt = transAmt.getAmount();
                //BigDecimal outRatio = CalcUtil.divide(outAmt, totalBal);
                Pair<String, BigDecimal> pair = Pair.of(loanAstNo, outAmt);
                outRatioPairs.add(pair);
            });
            Collections.sort(outRatioPairs, Comparator.comparing(Pair::getRight));
            for (Pair<String, BigDecimal> pair : outRatioPairs) {
                Money totalBal = loanAstMap.get(pair.getKey()).getPrinBal();
                Money outAmt = transAmtMap.get(pair.getKey());

                Money entryAmt = entryAmt(remain, remain, totalBal.subtract(outAmt));
                outAmt.addTo(entryAmt);
                if (remain.isZero()) {
                    break;
                }
            }
        }
        // 4. 校验
        VarChecker.checkZero(remain,
            "资产[{0}-{1}]流出处理：有[{2}]未分配完全 - [{3}]",
            instalDomain.getAstNo(), instalDomain.getTermNo(),
            remain, prinAmt);
        Money transDetailAmtSum = new Money();
        transAmtMap.forEach((loanAstNo,outAmt)->{
            transDetailAmtSum.addTo(outAmt);
        });
        VarChecker.checkEquals(transDetailAmtSum, prinAmt,
            "资产[{0}-{1}]流出处理：本金分配汇总[{2}]和待分配金额[{3}]不一致",
            instalDomain.getAstNo(), instalDomain.getTermNo(),
            transDetailAmtSum, prinAmt);
        return transAmtMap;
    }

    /**
     * 利息分摊
     *
     * @param instalTrans
     * @param transAmt
     * @param fundAmt
     * @param eventCode
     */
    private void repoIntr(LoanInstalTransDetail instalTrans, DebtMoney transAmt, DebtMoney fundAmt, EventCodeEnum eventCode, boolean forceSplit) {
        if (transAmt.isZero()) {
            return;
        }
        LoanInstalDomain instalDomain = instalTrans.getLoanInstal();

        // 获取分期下，从该分期分配出去的机构收益组合
        List<LoanAstProfit> loanAstProfits = instalDomain.getSourcedProfits();

        // 按最近变更日期排序：如果变更日期未更新，且还有余额，说明之前发生过转让
        Collections.sort(loanAstProfits, Comparator.comparing(LoanAstProfit::getLastUpdateDate));

        Date curProcessDate = null;
        List<LoanAstProfit> sameDateGroup = Lists.newArrayList();

        Map<String, LoanAstProfitTransDetail> astProfitTransDetails = Maps.newHashMap();
        DebtMoney remain = transAmt.cloneAll(DebtMoneyFieldEnum.PRIN);
        DebtMoney remainFundAmt = fundAmt.cloneAll(DebtMoneyFieldEnum.PRIN);

        for (LoanAstProfit astProfit : loanAstProfits) {
            if (astProfit.getBal().getSum().isZero()) {
                continue;
            }

            if (null == curProcessDate) {
                curProcessDate = astProfit.getLastUpdateDate();
                sameDateGroup.add(astProfit);
                continue;
            }

            if (forceSplit || DateTimeUtil.isSameDay(curProcessDate, astProfit.getLastUpdateDate())) {
                sameDateGroup.add(astProfit);
                continue;
            } else {
                entryIntrGroup(instalTrans, eventCode, astProfitTransDetails, sameDateGroup, remain, remainFundAmt,
                    transAmt, fundAmt);
            }
            curProcessDate = astProfit.getLastUpdateDate();
            sameDateGroup = Lists.newArrayList();
            sameDateGroup.add(astProfit);
        }
        entryIntrGroup(instalTrans, eventCode, astProfitTransDetails, sameDateGroup, remain, remainFundAmt, transAmt,
            fundAmt);

        // 3. 尾数处理
        for (DebtMoneyFieldEnum fieldType : ProfitMeasure.PFT_AMT_TYPES) {
            if (remain.getByFieldType(fieldType).isGreaterThanZero() || remainFundAmt.getByFieldType(fieldType)
                .isGreaterThanZero()) {
                BigDecimal fundRatio = calcFundRatio(fundAmt.getByFieldType(fieldType),
                    transAmt.getByFieldType(fieldType));
                processIntrTail(instalTrans, eventCode, astProfitTransDetails, loanAstProfits, remain,
                    remainFundAmt, fundRatio, fieldType);
            }
        }

        // 4. 校验
        remain.checkZero(MessageUtil.formatMsg("资产[{0}-{1}]利息回收处理：有[{2}]未分配完全 - [{3}]",
            instalTrans.getLoanInstal().getAstNo(), instalTrans.getLoanInstal().getAstNo(),
            remain, transAmt));
        remainFundAmt.checkZero(MessageUtil.formatMsg("资产[{0}-{1}]利息实收处理：有[{2}]未分配完全 - [{3}]",
            instalTrans.getLoanInstal().getAstNo(), instalTrans.getLoanInstal().getAstNo(),
            remainFundAmt, fundAmt));

        DebtMoney checkAmt = new DebtMoney();
        DebtMoney checkFundAmt = new DebtMoney();
        for (LoanAstProfitTransDetail transDetail : astProfitTransDetails.values()) {
            LoanAstProfit astProfit = transDetail.getLoanAstProfit();
            LoanAstProfitLog astProfitLog = transDetail.getActiveLog();
            checkAmt.addAll(astProfitLog.getAmt());
            checkFundAmt.addAll(astProfitLog.getFundAmt());

            if (eventCode.getSceneCode() == TransCodeEnum.REPO) {
                if (checkIsClear(astProfit.getOwnOrg(), astProfitLog.getAccountDate(),
                    astProfit.getRefLoanAst().getProdNo())) {
                    astProfitLog.getClrgAmt().addTo(transDetail.getActiveLog().getFundAmt().getSum());
                }
            }
        }

        checkAmt.setByFieldType(DebtMoneyFieldEnum.PRIN, transAmt.getByFieldType(DebtMoneyFieldEnum.PRIN));
        checkFundAmt.setByFieldType(DebtMoneyFieldEnum.PRIN, fundAmt.getByFieldType(DebtMoneyFieldEnum.PRIN));
        VarChecker.checkEquals(checkAmt, transAmt, "资产[{0}-{1}]流出处理：收益应收分配汇总[{2}]和待分配金额[{3}]不一致",
            instalTrans.getLoanInstal().getAstNo(), instalTrans.getLoanInstal().getAstNo(), checkAmt, transAmt);
        VarChecker.checkEquals(checkFundAmt, fundAmt, "资产[{0}-{1}]流出处理：收益实收分配汇总[{2}]和待分配金额[{3}]不一致",
            instalTrans.getLoanInstal().getAstNo(), instalTrans.getLoanInstal().getAstNo(), checkFundAmt, fundAmt);

    }

    private void entryIntrGroup(LoanInstalTransDetail instalTrans, EventCodeEnum eventCode,
                                Map<String, LoanAstProfitTransDetail> astProfitTransDetails,
                                List<LoanAstProfit> astProfits, DebtMoney remainTransAmt, DebtMoney remainFundAmt,
                                DebtMoney totalAmt, DebtMoney totalFundAmt) {

        for (DebtMoneyFieldEnum fieldType : ProfitMeasure.PFT_AMT_TYPES) {
            if (remainTransAmt.getByFieldType(fieldType).isGreaterThanZero()) {
                entryIntrField(instalTrans, eventCode, astProfitTransDetails, astProfits, remainTransAmt, remainFundAmt,
                    totalAmt, totalFundAmt, fieldType);
            }
        }
    }

    private void entryIntrField(LoanInstalTransDetail instalTrans, EventCodeEnum eventCode,
                                Map<String, LoanAstProfitTransDetail> astProfitTransDetails,
                                List<LoanAstProfit> astProfits, DebtMoney remain, DebtMoney remainFund,
                                DebtMoney total, DebtMoney totalFund, DebtMoneyFieldEnum fieldType) {

        Map<String, BigDecimal> ratios = Maps.newHashMap();
        Map<String, BigDecimal> dist = Maps.newHashMap();

        // 1. 构造利息在各个机构收益中分配比例
        Money totalBal = new Money();
        for (LoanAstProfit astProfit : astProfits) {
            dist.put(astProfit.getAstProfitNo(), astProfit.getBal().getByFieldType(fieldType).getAmount());
            totalBal.addTo(astProfit.getBal().getByFieldType(fieldType));
        }

        // 剩余金额足够分摊
        Money remainAmt = remain.getByFieldTypeWithoutClone(fieldType);
        Money remainFundAmt = remainFund.getByFieldTypeWithoutClone(fieldType);
        Money totalAmt = total.getByFieldType(fieldType);
        Money totalFundAmt = totalFund.getByFieldType(fieldType);
        BigDecimal fundRatio = calcFundRatio(totalFundAmt, totalAmt);

        if (MoneyUtil.compare(totalBal, remainAmt) <= 0) {
            // 足够分摊，就不计算本金了，直接分摊
            for (LoanAstProfit astProfit : astProfits) {
                LoanAstProfitTransDetail transDtl = getTransDetail(astProfitTransDetails, astProfit, instalTrans,
                    eventCode);
                Money bal = astProfit.getBal().getByFieldType(fieldType);
                Money entryAmt = entryAmt(remainAmt, bal, bal);
                transDtl.processTransAmt(entryAmt, fieldType);

                Money entryFundAmt = entryAmt.multiply(fundRatio);
                entryFundAmt = entryAmt(remainFundAmt, entryFundAmt);
                transDtl.processTransFundAmt(entryFundAmt, fieldType);
            }
        } else {
            DecimalCalcUtil.calcSplitFactor(ratios, dist, false);

            // 2. 计算各机构资产本金摊销分配
            Collections.sort(astProfits,
                (o1, o2) -> o2.getBal().getByFieldType(fieldType).compareTo(o1.getBal().getByFieldType(fieldType)));
            for (LoanAstProfit astProfit : astProfits) {

                // 2.1 构造交易明细
                LoanAstProfitTransDetail transDtl = getTransDetail(astProfitTransDetails, astProfit, instalTrans,
                    eventCode);

                // 2.2 入账
                BigDecimal ratio = ratios.get(astProfit.getAstProfitNo());
                Money entryAmt = CalcUtil.calcMoneyRatio(totalAmt, ratio);
                Money bal = astProfit.getBal().getByFieldType(fieldType);
                entryAmt = entryAmt(remainAmt, entryAmt, bal);
                transDtl.processTransAmt(entryAmt, fieldType);

                Money entryFundAmt = entryAmt.multiply(fundRatio);
                entryFundAmt = entryAmt(remainFundAmt, entryFundAmt);
                transDtl.processTransFundAmt(entryFundAmt, fieldType);

                if (remainAmt.isZero()) {
                    break;
                }
            }
        }

        // 因为是分组入账，因此不能再这里处理尾数。
    }

    /**
     * 利息尾数处理
     * @param instalTrans
     * @param eventCode
     * @param astProfitTransDetails
     * @param astProfits
     * @param remain
     * @param remainFund
     * @param fundRatio
     * @param fieldType
     */
    private void processIntrTail(LoanInstalTransDetail instalTrans, EventCodeEnum eventCode,
                                 Map<String, LoanAstProfitTransDetail> astProfitTransDetails,
                                 List<LoanAstProfit> astProfits, DebtMoney remain, DebtMoney remainFund,
                                 BigDecimal fundRatio, DebtMoneyFieldEnum fieldType) {
        Money remainAmt = remain.getByFieldTypeWithoutClone(fieldType);
        Money remainFundAmt = remainFund.getByFieldTypeWithoutClone(fieldType);

        Map<String, LoanAstProfit> astProfitMap = astProfits.stream().collect(
            Collectors.toMap(LoanAstProfit::getAstProfitNo, item -> item));

        if (remainAmt.isGreaterThanZero()) {
            // 计算 本次流出/总流出前余额
            // 根据 本次流出/总流出前余额 的比例排序，越小说明本金还的越少，优先摊销
            List<Pair<String, BigDecimal>> outRatioPairs = Lists.newArrayList();

            // 如果有剩余金额，则所有有利息余额的机构收益都已经得到处理，肯定有LoanAstProfitTransDetail， 所以扫描交易明细即可
            for (LoanAstProfitTransDetail astProfitTransDetail : astProfitTransDetails.values()) {

                LoanAstProfitLog astProfitLog = astProfitTransDetail.getActiveLog();
                LoanAstProfit astProfit = astProfitTransDetail.getLoanAstProfit();

                DebtMoney totalOut = astProfitLog.getAmt();
                BigDecimal outAmt = totalOut.getByFieldType(fieldType).getAmount();
                BigDecimal totalBal = astProfit.getBal().getByFieldType(fieldType).getAmount().add(outAmt);

                BigDecimal outRatio = CalcUtil.divide(outAmt, totalBal);
                Pair<String, BigDecimal> pair = Pair.of(astProfit.getAstProfitNo(), outRatio);
                outRatioPairs.add(pair);
            }

            Collections.sort(outRatioPairs, new Comparator<Pair<String, BigDecimal>>() {
                @Override
                public int compare(Pair<String, BigDecimal> o1, Pair<String, BigDecimal> o2) {
                    if (BigDecimalUtil.isEQ(o1.getRight(), o2.getRight())) {
                        LoanAstProfit left1 = astProfitMap.get(o1.getLeft());
                        LoanAstProfit left2 = astProfitMap.get(o2.getLeft());
                        return left1.getTermNo() - left2.getTermNo();
                    } else {
                        return o1.getRight().compareTo(o2.getRight());
                    }
                }
            });

            for (Pair<String, BigDecimal> outRatioPair : outRatioPairs) {
                LoanAstProfit astProfit = astProfitMap.get(outRatioPair.getLeft());
                LoanAstProfitTransDetail transDtl = getTransDetail(astProfitTransDetails, astProfit, instalTrans,
                    eventCode);
                Money entryAmt = entryAmt(remainAmt, remainAmt, astProfit.getBal().getByFieldType(fieldType));
                transDtl.processTransAmt(entryAmt, fieldType);

                Money fundAmt = entryAmt.multiply(fundRatio);
                fundAmt = entryAmt(remainFundAmt, fundAmt);
                transDtl.processTransFundAmt(fundAmt, fieldType);
                if (remainAmt.isZero()) {
                    break;
                }
            }
        }

        if (remainFundAmt.isGreaterThanZero()) {
            // 计算 实还差值：机构收益分摊前余额 * （机构收益分摊比例 - 总比例），负数说明分的偏少，正数说明分的偏多
            // 根据 实还/已还 的比例排序，越小说明本金还的越少，优先摊销
            List<Pair<String, BigDecimal>> diffPairs = Lists.newArrayList();

            for (LoanAstProfitTransDetail astProfitTransDetail : astProfitTransDetails.values()) {

                LoanAstProfitLog astProfitLog = astProfitTransDetail.getActiveLog();
                Money amt = astProfitLog.getAmt().getByFieldType(fieldType);
                if (amt.isZero()) {
                    continue;
                }
                Money fundAmt = astProfitLog.getFundAmt().getByFieldType(fieldType);
                BigDecimal ratio = CalcUtil.divide(fundAmt.getAmount(), amt.getAmount());

                // 当前比例 - 总比例
                BigDecimal ratioDiff = ratio.subtract(fundRatio);
                BigDecimal diffAmt = amt.getAmount().multiply(ratioDiff);

                Pair<String, BigDecimal> pair = Pair.of(astProfitLog.getAstProfitNo(), diffAmt);
                diffPairs.add(pair);
            }

            Collections.sort(diffPairs, new Comparator<Pair<String, BigDecimal>>() {
                @Override
                public int compare(Pair<String, BigDecimal> o1, Pair<String, BigDecimal> o2) {
                    if (BigDecimalUtil.isEQ(o1.getRight(), o2.getRight())) {
                        LoanAstProfit left1 = astProfitMap.get(o1.getLeft());
                        LoanAstProfit left2 = astProfitMap.get(o2.getLeft());
                        return left1.getTermNo() - left2.getTermNo();
                    } else {
                        return o1.getRight().compareTo(o2.getRight());
                    }
                }
            });

            for (Pair<String, BigDecimal> diffPair : diffPairs) {
                LoanAstProfit astProfit = astProfitMap.get(diffPair.getLeft());
                LoanAstProfitTransDetail transDtl = getTransDetail(astProfitTransDetails, astProfit, instalTrans,
                    eventCode);

                LoanAstProfitLog astProfitLog = transDtl.getActiveLog();
                Money amt = astProfitLog.getAmt().getByFieldType(fieldType);
                Money fundAmt = astProfitLog.getFundAmt().getByFieldType(fieldType);

                // 最多不能超过diffAmt，不然实还比应还还多
                Money diffAmt = amt.subtract(fundAmt);

                BigDecimal diffVal = diffPair.getRight();
                Money entryFundAmt = null;
                if (BigDecimalUtil.isLT(diffVal, BigDecimal.ZERO)) {
                    // 小于0, 先按差数补齐，
                    Money tryEntryAmt = new Money(diffVal).multiply(-1);
                    entryFundAmt = entryAmt(remainFundAmt, tryEntryAmt, diffAmt);
                    if (entryFundAmt.isZero()) {
                        // 差数后也等于0，默认加1分钱
                        entryFundAmt = entryAmt(remainFundAmt, Money.createWithCent(1L), diffAmt);
                    }
                } else {
                    // 极端情况会进这个分支，加1分钱
                    entryFundAmt = entryAmt(remainFundAmt, Money.createWithCent(1L), diffAmt);
                }
                transDtl.processTransFundAmt(entryFundAmt, fieldType);
                if (remainFundAmt.isZero()) {
                    break;
                }

                // 保证应还 >= 实还
                VarChecker.checkArgument(MoneyUtil.compare(astProfitLog.getAmt().getByFieldType(fieldType),
                    astProfitLog.getFundAmt().getByFieldType(fieldType)) >= 0,
                    "分期[{0}-{1}]利息回收分摊，应还[{2}]不能小于实还[{3}]",
                    astProfitLog.getAstNo(), astProfitLog.getTermNo(),
                    astProfitLog.getAmt().getByFieldType(fieldType),
                    astProfitLog.getFundAmt().getByFieldType(fieldType));
            }
        }

        LoanInstalDomain loanInstal = instalTrans.getLoanInstal();
        VarChecker.checkArgument(remainAmt.isZero(), "分期[{0}-{1}]应付流出分摊，剩余金额[{2}],{3}", loanInstal.getAstNo(),
            loanInstal.getTermNo(), remainAmt,fieldType);
        VarChecker.checkArgument(remainFundAmt.isZero(), "分期[{0}-{1}]实付流出分摊，剩余金额[{2}],{3}", loanInstal.getAstNo(),
            loanInstal.getTermNo(), remainFundAmt,fieldType);
    }

    private LoanAstProfitTransDetail getTransDetail(Map<String, LoanAstProfitTransDetail> astProfitTransDetails,
                                                    LoanAstProfit astProfit, LoanInstalTransDetail instalTransDetail,
                                                    EventCodeEnum eventCode) {

        LoanAstProfitTransDetail ret = astProfitTransDetails.get(astProfit.getAstProfitNo());

        if (null == ret) {
            ret = LoanAstProfitTransDetail.createTransDtl(astProfit, instalTransDetail, eventCode);
            astProfitTransDetails.put(astProfit.getAstProfitNo(), ret);
        }
        return ret;
    }

    private boolean checkIsClear(String ownOrg, Date accountDate, String prodNo) {
        PlanBase plan = PlanContextHolder.getPlan(ownOrg);

        if (plan == null) {
            plan = planManager.getPlanFullByNo(ownOrg);
            PlanContextHolder.putPlan(plan);
        }

        if (plan.getPlanType() == PlanTypeEnum.PLAN) {
            return true;
        } else {
            OrganizationEntity orgEntity = (OrganizationEntity)plan;
            boolean isNeedClr = PlanHandler.calcNeedClr(orgEntity, accountDate);
            VarChecker.checkArgument(orgEntity.getProdCodeList().contains(prodNo),
                "还款处理：产品{0}未在计划{1}中配置,当前仅支持产品{2}", prodNo, orgEntity.getPlanNo(),
                orgEntity.getProdCodeList());
            return isNeedClr;
        }
    }

    private Money entryAmt(Money remain, Money entryAmt, Money... others) {
        Money realEntryAmt = MoneyUtil.min(entryAmt, remain);

        if (others != null) {
            for (Money other : others) {
                realEntryAmt = MoneyUtil.min(realEntryAmt, other);
            }
        }

        remain.subtractFrom(realEntryAmt);
        return realEntryAmt;
    }

    private BigDecimal calcFundRatio(Money fundAmt, Money transAmt) {
        return CalcUtil.divide(fundAmt.getAmount(), transAmt.getAmount());
    }
}