/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2020 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.rule.node;

import com.abssqr.plat.common.facade.base.DebtMoney;
import com.abssqr.plat.common.facade.enums.AstPackTradeTypeEnum;
import com.abssqr.plat.common.facade.enums.DebtMoneyFieldEnum;
import com.abssqr.plat.common.facade.model.rule.AstPriceRule;
import com.abssqr.plat.common.model.domain.ast.LoanAst;
import com.abssqr.plat.common.model.domain.ast.LoanAstProfit;
import com.abssqr.plat.common.model.domain.ast.LoanAstProfitGroup;
import com.abssqr.plat.common.model.domain.rule.LoanAstPriceCalcParam;
import com.abssqr.plat.common.model.domain.rule.LoanAstPriceResult;
import com.abssqr.plat.common.model.domain.rule.LoanAstProfitBalAndPrice;
import com.abssqr.plat.common.model.utils.AstPriceUtil;
import com.abssqr.plat.core.service.util.AstPriceCalcAmtUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Maps;
import org.springframework.stereotype.Component;

import java.util.Map;

/**
 * 回购定价计算器
 * @author hanfei
 * @version $Id: AstBuyBackPriceCalc.java, v 0.1 2020-02-08 10:31 PM hanfei Exp $ 
 */
@Component
public class AstBuyBackPriceCalc extends AbstractPriceBaseCalc<AstPriceRule> {
    
    /**
     * 回购资产定价计算
     * @param param
     * @return
     */
    @Override
    protected LoanAstPriceResult handle(LoanAstPriceCalcParam<AstPriceRule> param) {
        // 定价规则
        AstPriceRule priceRule  = AstPriceUtil.filterPrice(param.getTradePriceRules(), param.getCondition());
        VarChecker.checkNotNull(priceRule, "未找到匹配的定价规则[{0}]-[{1}]",
                param.getCondition(), param.getTradePriceRules());
        // 本金
        LoanAst loanAst = param.getLoanAst();
        // 组装所有余额
        DebtMoney allCurBal = new DebtMoney();
        // 放入本金
        allCurBal.addByFieldType(DebtMoneyFieldEnum.PRIN, loanAst.getPrinBal());
        // 组装所有余额
        DebtMoney allPrice = new DebtMoney();
        // 计算本金定价
        allPrice.addByFieldType(DebtMoneyFieldEnum.PRIN,
                AstPriceCalcAmtUtil.calcPrice(loanAst.getPrinBal(), priceRule.getTradeRate(DebtMoneyFieldEnum.PRIN)));
        // 机构资产收益余额及定价金额 map
        Map<String, LoanAstProfitBalAndPrice> profitBalAndPriceMap = Maps.newHashMap();
        // 根据机构资产当期所有收益进行定价计算
        for (LoanAstProfitGroup astProfitGroup : loanAst.getLoanAstProfits().values()) {
            // 获取持有收益
            LoanAstProfit astProfit = astProfitGroup.getOwnAstProfit();

            // 当前收益余额
            DebtMoney profitBal = astProfit.getBal().cloneAll();
    
            // 定价金额金额
            DebtMoney price = new DebtMoney();
            // 计算定价金额
            for (DebtMoneyFieldEnum debtMoneyField : DebtMoneyFieldEnum.values()) {
                Money curBalField = profitBal.getByFieldType(debtMoneyField);
                // 追加所有收益余额
                allCurBal.addByFieldType(debtMoneyField, curBalField);
                // 定价
                Money priceField = AstPriceCalcAmtUtil.calcPrice(curBalField, priceRule.getTradeRate(debtMoneyField));
                // 追加所有定价
                allPrice.addByFieldType(debtMoneyField, priceField);
                // 对应当前收益余额的定价
                price.setByFieldType(debtMoneyField, priceField);
            }
    
            // 记录余额与定价扩展字段
            LoanAstProfitBalAndPrice balAndPrice = new LoanAstProfitBalAndPrice(profitBal.cloneAll(), price.cloneAll());
            profitBalAndPriceMap.put(astProfit.getAstProfitNo(), balAndPrice);
        }
        
        // 定价计算完成
        LoanAstPriceResult result = new LoanAstPriceResult();
        result.setBal(allCurBal);
        result.setPrice(allPrice);
        result.setTradePriceRule(priceRule);
        result.setTransIn(param.getTransIn());
        result.setTransOut(param.getTransOut());
        result.setProfitBalAndPriceMap(profitBalAndPriceMap);
        return result;
    }
    
    /**
     * 回购场景
     * @param param
     * @return
     */
    @Override
    public boolean judge(LoanAstPriceCalcParam param) {
        return param.getAstTradeType() == AstPackTradeTypeEnum.BUY_BACK_BIZ;
    }
}
